<!--
@Copyright 2015 Naver Corp.
@Author Yeseul Park (yeseul.park@navercorp.com)
-->

<!-- Extend our site layout -->
{% extends "layout.html" %}

<!-- Parse a pig tuple being passed through in string format -->

{% block content %}
	
  <div class="container">
  {% if tasknames %}
  <h3 style="margin-top: 20px;">[ Task History ]</h3>
  <ul class="nav nav-pills">
  {% for item in tasknames %}
	  <li class="active">
 	  	<a style="margin: 3px;" href="/tasklist/{{ item }}">{{ item }}</a>
  	  </li>
  {% end %}
  </ul>
  {% end %}
  </div>

{% end %}
